(ns client
  (:require 
            [cljs-http.client :as http]
            [cljs.core.async :refer [<! chan go]]
            [goog.crypt.base64 :as base64]))

(def api-url "https://api.runops.io")

(defn extract-body
  "Returns the `:body` key of the `response`.
  It returns the `:body` key of the map for successful HTTP status codes.
  Error HTTP status codes displays the `:message` key to the user,
  and return `nil`."
  [response]
  (case (:status response)
    200 (:body response)
    201 (:body response)
    0 (js/console.log "Failed to connect to the server. Please check your network connection and try again.")
    (js/console.log (:message (:body response)))))

(defn authorization-value [key]
      (if (clojure.string/starts-with? key "Bearer")
        key
        (str "Basic " (base64/encodeString (str key ":")))))

;;;; Tasks

(defn get-task [{:keys [id key]}]
  (http/get (str api-url "/v1/tasks/" id)
            {:headers {"Authorization" (authorization-value key)}
             :with-credentials? false}))

(defn run-task [{:keys [id key]}]
  (http/post (str api-url "/v1/tasks/" id "/runs")
            {:headers {"Authorization" (authorization-value key)}
             :with-credentials? false}))

(defn get-task-logs [{:keys [id key]}]
  (http/get (str api-url "/v1/tasks/" id "/logs")
            {:headers {"Authorization" (authorization-value key)}
             :with-credentials? false
             }))

(defn list-tasks [{:keys [key]}]
  (http/get (str api-url "/v1/tasks")
            {:headers {"Authorization" (authorization-value key)}
             :with-credentials? false}))

(defn create-task [{:keys [key message type target script]}]
  (http/post (str api-url "/v1/tasks")
             {:headers {"Authorization" (authorization-value key)}
              :with-credentials? false
              :json-params {:message message
                            :type type
                            :target target
                            :script script}}))

;;;; Targets 

(defn get-target [{:keys [key name]}]
  (http/get (str api-url "/v1/targets/" name)
            {:headers {"Authorization" (authorization-value key)}
             :with-credentials? false}))

(defn list-targets [{:keys [key]}]
  (http/get (str api-url "/v1/targets")
            {:headers {"Authorization" (authorization-value key)}
             :with-credentials? false}))

(defn create-target [target]
  (http/post (str api-url "/v1/targets")
             {:headers {"Authorization" (authorization-value (:key target))}
              :with-credentials? false
              :json-params (dissoc target :key)}))

(defn update-target [target]
  (http/put (str api-url "/v1/targets")
            {:headers {"Authorization" (authorization-value (:key target))}
              :with-credentials? false
              :json-params (dissoc target :key)}))

;;;; Templates

(defn list-templates [{:keys [key]}]
  (http/get (str api-url "/v2/templates")
            {:headers {"Authorization" (authorization-value key)}
             :with-credentials? false}))

(defn run-template [template]
  (http/post (str api-url "/v1/templates/" (:name template) "/runs")
             {:headers {"Authorization" (authorization-value (:key template))}
              :with-credentials? false
              :json-params (dissoc template :key :name)}))

(defn get-template [{:keys [name key]}]
  (http/get (str api-url "/v1/templates/" name)
            {:headers {"Authorization" (authorization-value key)}
             :with-credentials? false}))

;;;; Users

(defn list-users [{:keys [key]}]
  (http/get (str api-url "/v1/users")
            {:headers {"Authorization" (authorization-value key)}
             :with-credentials? false}))

(defn create-user [user]
  (http/post (str api-url "/v1/users")
             {:headers {"Authorization" (authorization-value (:key user))}
              :with-credentials? false
              :json-params (dissoc user :key)}))

(defn update-user [user]
  (http/put (str api-url "/v1/users/" (:id user))
            {:headers {"Authorization" (authorization-value (:key user))}
             :with-credentials? false
             :json-params (dissoc user :id :key)}))
