(ns core
  (:require [cljs.nodejs :as node]
            [cljs.core.async :refer [<! chan go]]
            [xmlhttprequest :refer [XMLHttpRequest]]
            [client :as client]
            [targets :as targets]
            [users :as users]
            [templates :as templates]
            [traces :as traces]
            [auth :as auth]
            [login :as login]
            [spinner :as spinner]
            [cljs.core.async.interop :refer-macros [<p!]]))

(def program (node/require "commander"))
(def Table (node/require "cli-table"))

(defn node-slurp [path]
  (let [fs (node/require "fs")]
    (.readFileSync fs path "utf8")))

(defn extract-logs [task]
  (if (:task_logs task) (:task_logs task) task))

(defn extract-body
  "Returns the `:body` key of the `response`.
  It returns the `:body` key of the map for successful HTTP status codes.
  Error HTTP status codes displays the `:message` key to the user,
  and return `nil`."
  [response]
  (case (:status response)
    200 (:body response)
    201 (:body response)
    (js/console.log (:message (:body response)))))

(defn parse-content [body]
  (-> body
      prn
      with-out-str
      cljs.reader/read-string))

(defn print-table [table]
  (-> table .toString js/console.log))

(defn fill [table data]
  (doall (map #(.push table (clj->js %)) data)))

;;; List Tasks

(defn list-tasks-api [task]
  (let [_ (spinner/start)
        response-chan (client/list-tasks task)]
    (assoc task :response response-chan)))

(defn remove-nils [data]
  (into {} (map #(if (second %) % {(first %) ""}) data)))

(defn tasks-values [content]
  (->> content
       (map remove-nils)
       (map #(vals (select-keys % [:id :status :created :description])))))

(defn build-table [tasks]
  (let [table (Table. (clj->js {:head ["Id" "Status" "Created" "Message"]
                                :colWidths [7, 15, 20, 60]}))]
    (doto table (fill tasks))))

(defn list-tasks-result [task]
  (go (when-let [response (<! (:response task))]
        (some-> response
                spinner/stop
                extract-body
                parse-content
                tasks-values
                build-table
                print-table
                close))))

(defn list-tasks [list options]
  (.withTrace traces/event
              (clj->js {:command "list-tasks"})
              #(some-> (auth/key-auth {})
                       list-tasks-api
                       list-tasks-result)))

;;; Create Task

(defn read-script [task]
  (try (if (:script task)
         task
         (assoc task :script (node-slurp (:file task))))
       (catch js/Error e (js/console.log (.toString e)))))

(defn new-task-result [task]
  (go (when-let [response (<! (:response task))]
        (some-> response
                spinner/stop
                extract-body
                parse-content
                extract-logs
                clj->js
                js/console.log
                close))))

(defn new-task-api [task]
  (let [_ (spinner/start)
        response-chan (client/create-task task)]
    (assoc task :response response-chan)))

(defn create-task [command]
  (.withTrace traces/event (clj->js {:command "create-task"})
              #(some-> (conj {:target (.-target command)
                              :script (.-script command)
                              :file (.-file command)}
                             (when-let [message (.-message command)]
                               {:message message})
                             (when-let [task-type (.-type command)]
                               {:type task-type}))
                       read-script
                       auth/key-auth
                       new-task-api
                       new-task-result)))

;;; Get Task

(defn get-task-result [task]
  (go (when-let [response (<! (:response task))]
        (some-> response
                spinner/stop
                extract-body
                parse-content
                clj->js
                js/console.log
                close))))

(defn get-task-api [task]
  (let [_ (spinner/start)
        response-chan (client/get-task task)]
    (assoc task :response response-chan)))

(defn get-task [command]
  (.withTrace traces/event
              (clj->js {:command "get-task"})
              #(some-> (auth/key-auth {:id (.-id command)})
                       get-task-api
                       get-task-result)))

;;; Get Task Logs

(defn get-task-logs-result [task]
  (go (when-let [response (<! (:response task))]
        (some-> response
                spinner/stop
                extract-body
                clj->js
                js/console.log
                close))))

(defn get-task-logs-api [task]
  (let [_ (spinner/start)
        response-chan (client/get-task-logs task)]
    (assoc task :response response-chan)))

(defn get-task-logs [command]
  (.withTrace traces/event
              (clj->js {:command "get-task-logs"})
              #(some-> (auth/key-auth {:id (.-id command)})
                       get-task-logs-api
                       get-task-logs-result)))

;;; Run Task

(defn run-task-result [task]
  (go (when-let [response (<! (:response task))]
        (some-> response
                spinner/stop
                extract-body
                parse-content
                extract-logs
                clj->js
                js/console.log
                close))))

(defn run-task-api [task]
  (let [_ (spinner/start)
        response-chan (client/run-task task)]
    (assoc task :response response-chan)))

(defn run-task [command]
  (.withTrace traces/event
              (clj->js {:command "run-task"})
              #(some-> (auth/key-auth {:id (.-id command)})
                       run-task-api
                       run-task-result)))

;;;; CLI 

;;; Signup

(defn signup-commands [program]
  (let [signup (-> program (.command "signup"))]
    (doto signup
      (.description "Signup")
      (.action login/signup))))

;;; Login

(defn login-commands [program]
  (let [login (-> program (.command "login"))]
    (doto login
      (.description "Login")
      (.action login/login))))


(defn logout-commands [program]
  (let [logout (-> program (.command "logout"))]
    (doto logout
      (.description "logout")
      (.action login/logout))))

;;; Templates 

(defn templates-commands [program]
  (let [templates (-> program (.command "templates"))
        get       (.command templates "get")
        list      (.command templates "list")
        run       (.command templates "run")]
    (doto list
      (.description "List Templates")
      (.action templates/list-templates))

    (doto get
      (.description "Get Template Params")
      (.requiredOption "-n --name <name>" "The name of the Template")
      (.action templates/get-template))

    (doto run
      (.description "Create a Task using a Template")
      (.requiredOption "-n --name <name>" "The name of the Template")
      (.option "-m --message <message>" "A message with the description of the task")
      (.requiredOption "-t --target <target>" "Target where the task should run (ex: db-prod, k8s-homolog)")
      (.option "-y --type <type>" "The type of the script (ex: postgres, python, k8s) - overrides the Target type")
      (.option "-p --params <params>"
               "A comma-separated list with the parameters required in the Template, in the format: `param_name=param_value`")
      (.action templates/run-template))))

;;; Targets

(defn targets-commands [program]
  (let [targets (-> program (.command "targets"))
        create    (.command targets "create")
        update    (.command targets "update")
        list    (.command targets "list")
        get    (.command targets "get")]
    (doto list
      (.description "List Targets")
      (.action targets/list-targets))

    (doto get
      (.description "Get Target")
      (.requiredOption "-n --name <name>" "The name of the target")
      (.action targets/get-target))

    (doto create
      (.description "Create a new Target. Check out https://docs.runops.io")
      (.requiredOption "--name <name>"
                       "The name used to create Tasks with this Target")
      (.requiredOption "--type <type>"
                       "The type of the Target")
      (.option "--message <message>"
               "A message to help users identify the Target")
      (.option "--runner <runner>"
               "The type of runner for the Target: `runops-hosted` or `self-hosted`", "runops-hosted")
      (.option "--review <review>"
               "`none`: no reviews. `default`: the settings of the org. `teams`: the list of teams in the `reviewers` option")
      (.option "--channel <channel>"
               "The name of the Slack channel to send reviews")
      (.option "--reviewers <reviewers>"
               "A comma-separated list of names with the Teams that must review Tasks on this Target.")
      (.option "--secrets <secrets>"
               "A comma-separated list with the secrets to attach to the Target, in the format: `SECRET_NAME=SECRET_VALUE`")
      (.option "--runner_provider <message>"
               "The runner provider: `runops` or `github`. Defaults to `github` if not provided")
      (.option "--secret_provider <message>"
               "The secret manager provider: `aws`, `hashcorp` or `gcp`. Defaults to null if not provided")
      (.option "--secret_path <message>"
               "The secret name or path in the provider")
      (.option "--tags <message>"
               "For `runops` runner provider only, it identifies the environment. Any string like `dev`, `prod`, etc")
      (.option "--roles <message>"
               "A comma separated string with all allowed user roles that can read and use the target")
      (.option "--config <message>"
               "A JSON structure to config target")
      (.action targets/create))

    (doto update 
      (.description "Update a Target. Check out https://docs.runops.io")
      (.requiredOption "--name <name>"
                       "The name of the Target to update")
      (.option "--type <type>"
               "The type of the Target")
      (.option "--message <message>"
               "A message to help users identify the Target")
      (.option "--runner <runner>"
               "The type of runner for the Target: `runops-hosted` or `self-hosted`")
      (.option "--review <review>"
               "`none`: no reviews. `default`: the settings of the org. `teams`: the list of teams in the `reviewers` option")
      (.option "--channel <channel>"
               "The name of the Slack channel to send reviews")
      (.option "--reviewers <reviewers>"
               "A comma-separated list of names with the Teams that must review Tasks on this Target")
      (.option "--secrets <secrets>"
               "A comma-separated list with the secrets to attach to the Target, in the format: `SECRET_NAME=SECRET_VALUE`")
      (.option "--runner_provider <message>"
               "The runner provider: `runops` or `github`. Defaults to `github` if not provided")
      (.option "--secret_provider <message>"
               "The secret manager provider: `aws`, `hashcorp` or `gcp`. Defaults to null if not provided")
      (.option "--secret_path <message>"
               "The secret name or path in the provider")
      (.option "--tags <message>"
               "For `runops` runner provider only, it identifies the environment. Any string like `dev`, `prod`, etc")
      (.option "--roles <message>"
               "A comma separated string with all allowed user roles that can read and use the target")
      (.option "--config <message>"
               "A JSON structure to config target")
      (.action targets/update))))

;;; Tasks

(defn tasks-commands [program]
  (let [tasks (-> program (.command "tasks"))
        create (.command tasks "create")
        list (.command tasks "list")
        logs (.command tasks "logs")
        get (.command tasks "get")
        run (.command tasks "run")]

  (doto create
    (.description "Create a new Task. Check out https://docs.runops.io")
    (.option "-m --message <message>" "A message with the description of the task")
    (.requiredOption "-t --target <target>" "Target where the task should run 🚨-t is now the Target (previously -g)🚨")
    (.option "-y --type <type>" "The type of the script - overrides the Target type")
    (.option "-s --script <script>" "The contents of the script of the task")
    (.option "-f --file <file>" "The path to a local file with the script")
    (.action create-task))

  (doto list
    (.description "List tasks")
    (.action list-tasks))

  (doto get
    (.description "Get task details")
    (.requiredOption "-i --id <id>" "The id of the task")
    (.action get-task))

  (doto run
    (.description "Run a Task")
    (.requiredOption "-i --id <id>" "The id of the task")
    (.action run-task))

  (doto logs
    (.description "Get task logs")
    (.requiredOption "-i --id <id>" "The id of the task")
    (.action get-task-logs))))

;;; Users

(defn users-commands [program]
  (let [users (-> program (.command "users"))
        create (.command users "create")
        update (.command users "update")
        list (.command users "list")]
    (doto list
      (.description "List Users")
      (.action users/list-users))

    (doto create
      (.description "Create a User")
      (.requiredOption "--name <name>"
                       "The name of the User")
      (.requiredOption "--email <email>"
                       "The email of the User")
      (.option "--status <status>"
               "The status of the User")
      (.option "--team <team>"
               "The team of the User")
      (.option "--roles <roles>"
               "A comma separated string with user roles")
      (.action users/create))

    (doto update
      (.description "Update a User")
      (.requiredOption "-i --id <id>"
                       "The id of the User to update")
      (.option "--status <status>"
               "The status of the User")
      (.option "--team <team>"
               "The team of the User")
      (.option "--roles <roles>"
               "A comma separated string with user roles")
      (.action users/update))))

(let [_ (tasks-commands program)
      _ (targets-commands program)
      _ (templates-commands program)
      _ (login-commands program)
      _ (logout-commands program)
      _ (signup-commands program)
      _ (users-commands program)])

(defn start [& _]
  (doto program
    (.version "0.3.7")
    (.parse process.argv)))

(node/enable-util-print!)
(set! js/XMLHttpRequest XMLHttpRequest)
(set! *main-cli-fn* start)

(defn main [& cli-args]
  (start))
