(ns targets
  (:require [cljs.nodejs :as node]
            [cljs.core.async :refer [<! chan go]]
            [client :as client]
            [traces :as traces]
            [auth :as auth]
            [spinner :as spinner]
            [cljs.core.async.interop :refer-macros [<p!]]))

(defn parse-content [body]
  (-> body
      prn
      with-out-str
      cljs.reader/read-string))

(defn key-value-str->map [data]
  (let [index (clojure.string/index-of data "=" 0)]
    {:name (subs data 0 index)
     :value (subs data (+ index 1))}))

(defn parse-secrets [target]
  (if-let [secrets (:secrets target)]
    (let [secrets (clojure.string/split secrets #",")
          secrets (map key-value-str->map secrets)]
      (assoc target :secrets secrets))
    target))

(defn parse-roles [target]
  (if-let [roles (:roles target)]
    (let [roles (clojure.string/split roles #",")]
      (assoc target :roles roles))
    target))

(defn create-target-api [target]
  (let [_ (spinner/start)
        response-chan (client/create-target target)]
    (assoc task :response response-chan)))

(defn update-target-api [target]
  (let [_ (spinner/start)
        response-chan (client/update-target target)]
    (assoc task :response response-chan)))

(defn flat-target-fields [target]
  (assoc target :roles (clojure.string/join "," (:roles target))))

(defn get-target-result [target]
  (go (when-let [response (<! (:response target))]
        (some-> response
                spinner/stop
                client/extract-body
                parse-content
                flat-target-fields
                clj->js
                js/console.log
                traces/close))))

(defn create [command]
  (.withTrace traces/event (clj->js {:command "create-target"})
              #(some-> {:name (.-name command)
                        :type (.-type command)
                        :message (.-message command)
                        :runner (.-runner command)
                        :review_type (.-review command)
                        :reviewers (.-reviewers command)
                        :channel_name (.-channel command)
                        :runner_provider (.-runner_provider command)
                        :secret_provider (.-secret_provider command)
                        :secret_path (.-secret_path command)
                        :tags (.-tags command)
                        :roles (.-roles command)
                        :config (.-config command)
                        :secrets (.-secrets command)}
                       parse-secrets
                       parse-roles
                       auth/key-auth
                       create-target-api
                       get-target-result)))

(defn update [command]
  (.withTrace traces/event (clj->js {:command "update-target"})
              #(some-> (conj {:name (.-name command)}
                             (when-let [type (.-type command)]
                               {:type type})
                             (when-let [message (.-message command)]
                               {:message message})
                             (when-let [runner (.-runner command)]
                               {:runner runner})
                             (when-let [review (.-review command)]
                               {:review_type review})
                             (when-let [reviewers (.-reviewers command)]
                               {:reviewers reviewers})
                             (when-let [channel (.-channel command)]
                               {:channel_name channel})
                             (when-let [secrets (.-secrets command)]
                               {:secrets secrets})
                             (when-let [runner_provider (.-runner_provider command)]
                               {:runner_provider runner_provider})
                             (when-let [secret_provider (.-secret_provider command)]
                               {:secret_provider secret_provider})
                             (when-let [secret_path (.-secret_path command)]
                               {:secret_path secret_path})
                             (when-let [config (.-config command)]
                               {:config config})
                             (when-let [tags (.-tags command)]
                               {:tags tags})
                             (when-let [roles (.-roles command)]
                               {:roles roles}))
                       (parse-secrets)
                       (parse-roles)
                       (auth/key-auth)
                       (update-target-api)
                       (get-target-result))))

;;; List Targets 

(def Table (node/require "cli-table"))

(defn list-targets-api [target]
  (let [_ (spinner/start)
        response-chan (client/list-targets target)]
    (assoc target :response response-chan)))

(defn get-target-api [target]
  (let [_ (spinner/start)
        response-chan (client/get-target target)]
    (assoc target :response response-chan)))

(defn remove-nils [data]
  (into {} (map #(if (second %) % {(first %) ""}) data)))

(defn targets-values [content]
  (->> content
       (map remove-nils)
       (map #(vals (select-keys % [:name :type :review_type :reviewers :tags])))))

(defn fill [table data]
  (doall (map #(.push table (clj->js %)) data)))

(defn build-targets-table [targets]
  (let [table (Table. (clj->js {:head ["name" "type" "review" "reviewers" "tags"]
                                :colWidths [25, 10, 10, 25, 10]}))
        table (doto table (fill targets))]
    table))

(defn print-table [table]
  (-> table .toString js/console.log))

(defn list-targets-result [target]
  (go (when-let [response (<! (:response target))]
        (some-> response
                spinner/stop
                client/extract-body
                parse-content
                targets-values
                build-targets-table
                print-table
                traces/close))))

(defn list-targets [list options]
  (.withTrace traces/event
              (clj->js {:command "list-targets"})
              #(some-> (auth/key-auth {})
                       list-targets-api
                       list-targets-result)))

(defn get-target [command]
  (.withTrace traces/event
              (clj->js {:command "get-target"})
              #(some-> (auth/key-auth {:name (.-name command)})
                       get-target-api
                       get-target-result)))
