(ns templates
  (:require [cljs.nodejs :as node]
            [cljs.core.async :refer [<! chan go]]
            [client :as client]
            [traces :as traces]
            [auth :as auth]
            [spinner :as spinner]
            [cljs.core.async.interop :refer-macros [<p!]]))

(defn parse-content [body]
  (-> body
      prn
      with-out-str
      cljs.reader/read-string))


;;; List Templates 

(def Table (node/require "cli-table"))

(defn list-templates-api [user]
  (let [_ (spinner/start)
        response-chan (client/list-templates user)]
    (assoc user :response response-chan)))

(defn templates-params-values [content]
  (map #(vals (select-keys % [:name :params])) content))

(defn templates-values [content]
  (map #(vals (select-keys % [:name])) content))

(defn format-params [templates]
  (map #(assoc % :params (clojure.string/join "\n" (:params %))) templates))

(defn names-to-map [templates]
  (map #(assoc {} :name %) templates))

(defn fill [table data]
  (doall (map #(.push table (clj->js %)) data)))

(defn build-templates-table [templates]
  (let [table (Table. (clj->js {:head ["name" "params"]
                                :colWidths [30, 40]}))
        table (doto table (fill templates))]
    table))

(defn build-templates-names-table [templates]
  (let [table (Table. (clj->js {:head ["name"]
                                :colWidths [30]}))
        table (doto table (fill templates))]
    table))

(defn print-table [table]
  (-> table .toString js/console.log))

(defn list-templates-result [template]
  (go (when-let [response (<! (:response template))]
        (some-> response
                spinner/stop
                client/extract-body
                names-to-map
                templates-values
                build-templates-names-table
                print-table
                traces/close))))

(defn list-templates [list options]
  (.withTrace traces/event
              (clj->js {:command "list-templates"})
              #(some-> (auth/key-auth {})
                       list-templates-api
                       list-templates-result)))


;;; Get Template
(defn conj-to-list [template]
  [template])

(defn get-template-result [template]
  (go (when-let [response (<! (:response template))]
        (some-> response
                spinner/stop
                client/extract-body
                conj-to-list
                parse-content
                format-params
                templates-params-values
                build-templates-table
                print-table
                traces/close))))

(defn get-template-api [template]
  (let [_ (spinner/start)
        response-chan (client/get-template template)]
    (assoc template :response response-chan)))

(defn get-template [command options]
  (.withTrace traces/event
              (clj->js {:command "get-template"})
              #(some-> (auth/key-auth {:name (.-name command)})
                       get-template-api
                       get-template-result)))


;;; Run templates
(defn run-template-api [template]
  (let [_ (spinner/start)
        response-chan (client/run-template template)]
    (assoc template :response response-chan)))

(defn run-template-result [template]
  (go (when-let [response (<! (:response template))]
        (some-> response
                spinner/stop
                client/extract-body
                clj->js
                js/console.log
                traces/close))))

(defn params-str->map [params param]
  (let [index (clojure.string/index-of param "=" 0)]
    (conj params
          {(keyword (subs param 0 index)) (subs param (+ index 1))})))

(defn parse-params [template]
  (if-let [params (:params template)]
    (let [params (clojure.string/split params #",")
          params (reduce params-str->map {} params)]
      (assoc template :params params))
    template))

(defn run-template [run options]
  (.withTrace traces/event
              (clj->js {:command "run-template"})
              #(some-> (conj {:name (.-name run)
                              :target (.-target run)}
                             (when-let [message (.-message run)]
                               {:message message})
                             (when-let [task-type (.-type run)]
                               {:type task-type})
                             (when-let [params (.-params run)]
                               {:params params}))
                       parse-params
                       auth/key-auth
                       run-template-api
                       run-template-result)))
