(ns users
  (:require [cljs.nodejs :as node]
            [cljs.core.async :refer [<! chan go]]
            [client :as client]
            [traces :as traces]
            [auth :as auth]
            [spinner :as spinner]
            [cljs.core.async.interop :refer-macros [<p!]]))

;; List User

(def Table (node/require "cli-table"))

(defn parse-content [body]
  (-> body
      prn
      with-out-str
      cljs.reader/read-string))

(defn remove-nils [data]
  (into {} (map #(if (second %) % {(first %) ""}) data)))

(defn users-values [content]
  (->> content
       (map remove-nils)
       (map #(vals (select-keys % [:id :name :email :status :team :roles])))))

(defn fill [table data]
  (doall (map #(.push table (clj->js %)) data)))

(defn build-users-table [users]
  (let [table (Table. (clj->js {:head ["id" "name" "email" "status" "team" "roles"]
                                :colWidths [5, 10, 25, 10, 10, 20]}))
        table (doto table (fill users))]
    table))

(defn print-table [table]
  (-> table .toString js/console.log))

(defn list-users-api [users]
  (let [_ (spinner/start)
        response-chan (client/list-users users)]
    (assoc users :response response-chan)))

(defn list-users-result [target]
  (go (when-let [response (<! (:response target))]
        (some-> response
                spinner/stop
                client/extract-body
                parse-content
                users-values
                build-users-table
                print-table
                traces/close))))

(defn list-users []
  (.withTrace traces/event
              (clj->js {:command "list-users"})
              #(some-> (auth/key-auth {})
                       list-users-api
                       list-users-result)))

;; Create/Update User
(defn create-user-api [user]
  (let [_ (spinner/start)
        response-chan (client/create-user user)]
    (assoc user :response response-chan)))

(defn update-user-api [user]
  (let [_ (spinner/start)
        response-chan (client/update-user user)]
    (assoc user :response response-chan)))

(defn parse-roles [user]
  (if-let [roles (:roles user)]
    (let [roles (clojure.string/split roles #",")]
      (assoc user :roles roles))
    user))

(defn flat-user-fields [user]
  (assoc user :roles (clojure.string/join "," (:roles user))))

(defn get-user-result [user]
  (go (when-let [response (<! (:response user))]
        (some-> response
                spinner/stop
                client/extract-body
                parse-content
                flat-user-fields
                clj->js
                js/console.log
                traces/close))))

(defn create [command]
  (.withTrace traces/event (clj->js {:command "create-user"})
              #(some-> (conj {:name (.-name command)
                              :email (.-email command)}
                             (when-let [status (.-status command)]
                               {:status status})
                             (when-let [team (.-team command)]
                               {:team team})
                             (when-let [roles (.-roles command)]
                               {:roles roles}))
                       (parse-roles)
                       (auth/key-auth)
                       (create-user-api)
                       (get-user-result))))

(defn update [command]
  (.withTrace traces/event (clj->js {:command "update-user"})
              #(some-> (conj {:id (.-id command)}
                             (when-let [status (.-status command)]
                               {:status status})
                             (when-let [team (.-team command)]
                               {:team team})
                             (when-let [roles (.-roles command)]
                               {:roles roles}))
                       (parse-roles)
                       (auth/key-auth)
                       (update-user-api)
                       (get-user-result))))